import logging
from typing import Dict, Union
from pathlib import Path
import yaml
from research_utils.decorators.logging import logger

def str_to_class(module_name: str, class_name: str):
  import importlib
  module = importlib.import_module(module_name)
  return getattr(module, class_name)

@logger()
def load_yaml(path: Union[str, Path]) -> Dict:
  with open(path, "r") as stream:
    try:
      return yaml.safe_load(stream)
    except yaml.YAMLError as e:
      logging.error(e)


from argparse import Action, ArgumentParser, Namespace
from typing import Any, Callable, Sequence, Tuple, Union


class YAMLAction(Action):
  """Argparser Action for Handling YAML file
  """
  def __init__(self, transform: Callable[[str], Union[Tuple[str, Any], Sequence[Tuple[str, Any]]]]=None, **kwargs) -> None:
    super().__init__(**kwargs)
    self._transform = transform

  def __call__(self,
               parser: ArgumentParser,
               namespace: Namespace,
               values: Union[str, Sequence[Any]],
               option_string: str = None) -> None:
    assert isinstance(values, str)
    if self._transform:
      delattr(namespace, self.dest)
      values = self._transform(values)
      if isinstance(values, Tuple):
        k, v = values
        setattr(namespace, k, v)
      elif isinstance(values, Sequence):
        for value in values:
          k, v = value
          setattr(namespace, k, v)
    else:
      setattr(namespace, self.dest, values)



from .sql import SQLArgument
from functools import reduce


class OperatorSQLArgument(SQLArgument):
  def __init__(self, field, operator=None, bound=None):
    super().__init__()
    self.__field = field
    self.__operator = operator
    self.__bound = bound

  @property
  def sql(self):
    if self.__operator is None and self.__bound is not None:
      return
    else:
      return f'{self.__field} {self.__operator} {self.__bound}'


class ConditionSQLArgument(SQLArgument):
  def __init__(self, *args, keyword=None):
    super().__init__()
    self.__keyword = keyword
    self.__args = args
    assert len(self.__args) > 0

  @property
  def sql(self):
    if self.__keyword is None:
      return
    elif len(self.__args) == 1:
      return self.__args[0].sql
    else:
      return reduce(lambda arg1, arg2: f'{arg1.sql} {self.__keyword} {arg2.sql}', self.__args)


class AND(ConditionSQLArgument):
  def __init__(self, *args):
    super().__init__(*args, keyword='AND')


class Lower(OperatorSQLArgument):
  def __init__(self, field, bound):
    super().__init__(field, operator='<=', bound=bound)


class Upper(OperatorSQLArgument):
  def __init__(self, field, bound):
    super().__init__(field, operator='>=', bound=bound)


class In(OperatorSQLArgument):
  def __init__(self, field, bound=None):
    super().__init__(field, operator='IN', bound=tuple(bound))


class NotIn(OperatorSQLArgument):
  def __init__(self, field, bound=None):
    super().__init__(field, operator='NOT IN', bound=tuple(bound))


class SQLArgumentFactory:
  @classmethod
  def between(self, field, lower_bound, upper_bound):
    return AND(Lower(field, upper_bound), Upper(field, lower_bound))



from typing import Any, Callable, Dict, List, Tuple, Union
import logging
from ..decorators.logging import logger
from pathlib import Path
import sqlite3
from sqlite3 import Error, Connection, Cursor
from .typing.sql import FieldOption
from .typing.operator import ConditionSQLArgument, OperatorSQLArgument


def create_db_if_not_exist(database) -> None:
  file = Path(database)
  file.parents[0].mkdir(parents=True, exist_ok=True)
  file.touch(exist_ok=True)
  logging.debug(f'{file} is created')


@logger()
def create_connection(database=':memory:', *args, **kwargs) -> Connection:
  """Create a Database Connection
  Args:
      database: {StrOrBytesPath} -- use ":memory:" to open a database 
          connection to a database that resides in RAM instead of on disk
      check_same_thread: {bool} -- If set False, the returned connection 
          may be shared across multiple threads {default: {True}}
  Returns:
      Connection: ...
  """

  if database != ':memory:':
    create_db_if_not_exist(database)

  conn = None
  try:
    conn = sqlite3.connect(database, *args, **kwargs)
  except Error as e:
    logging.error(e)
  return conn


def create_table(conn: Connection, table_name: str, fields: Dict[str, FieldOption]):
  """Create Table

  Args:
      conn (Connection): The Connection object
      table_name (str): Name of table
      fields (Dict[str, FieldOption]): Dictionary describing information of the fields
  """
  sql = ''.join((f'CREATE TABLE IF NOT EXISTS {table_name} (',
                 *(f'{key} {value.sql}{"," if index != len(fields) - 1 else ""}\n'
                   for index, (key, value) in enumerate(fields.items())), ');'))
  logging.info(sql)
  try:
    cur = conn.cursor()
    cur.execute(sql)
  except Error as e:
    logging.error(e)


def insert(conn: Connection, table_name: str, fields: Union[List[str], Tuple[str]],
           values: Union[List[Tuple], Tuple]):
  """Create Row or Rows into Custom Table

  Args:
      conn (Connection): The Connection object
      table_name (str): Name of table
      fields (Union[List[str], Tuple[str]]): 
      values (Union[List[Tuple], Tuple]): 
  """
  sql = f"""INSERT INTO {table_name} ({', '.join(map(str, fields))})
  VALUES
  ({', '.join(['?' for _ in fields])})
  """
  logging.info(sql)
  try:
    cur = conn.cursor()
    if isinstance(values, tuple):
      cur.execute(sql, values)
    else:
      cur.executemany(sql, values)
    conn.commit()
  except Error as e:
    logging.error(e)


@logger()
def query(conn: Connection,
          table_name: str,
          fields: Union[List[str], Tuple[str]] = None,
          row_factory: Callable[[Cursor, Tuple], Any] = None,
          where: Union[OperatorSQLArgument, ConditionSQLArgument] = None) -> List:
  """Query All Rows from Custom Table

  Args:
      conn (Connection): The Connection object
      table_name (str): Name of table
      fields (Union[List[str], Tuple[str]], optional): Defaults to None.
      where (Union[OperatorSQLArgument, ConditionSQLArgument], optional): Defaults to None.

  Returns:
      List: List of rows
  """

  sql = f"""SELECT {', '.join(map(str, fields)) if fields is not None and len(fields) > 0 else '*'} FROM {table_name}
  {f'WHERE {where.sql}' if where is not None else ''}
  """
  logging.info(sql)

  if row_factory is not None:
    conn.row_factory = row_factory

  try:
    cur = conn.cursor()
    cur.execute(sql)
    rows = cur.fetchall()
    return rows
  except Error as e:
    logging.error(e)



from typing import Callable
import inspect
import functools

def ignore_unknown_kwargs(func: Callable):
  """Ignore unknown kwargs

  Args:
    func (Callable): 

  If the function already has the catch all **kwargs, do nothing.
  """
  if any(param.kind == inspect.Parameter.VAR_KEYWORD for param in inspect.signature(func).parameters.values()):
    return func

  @functools.wraps(func)
  def inner(*args, **kwargs):
    filtered_kwargs = {
      name: kwargs[name]
      for name, param in inspect.signature(func).parameters.items() if (
        param.kind is inspect.Parameter.KEYWORD_ONLY or
        param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD
      ) and name in kwargs
    }
    return func(*args, **filtered_kwargs)
  return inner


from enum import Enum
from typing import Dict, Sequence, Union
from research_utils.sqlite.typing.sql import FieldType
from research_utils.utils import str_to_class


class SQLArgumentEnum(Enum):
  AND = ('research_utils.sqlite.typing.operator', 'AND')
  BETWEEN = ('research_utils.sqlite.typing.operator', 'SQLArgumentFactory', 'between')
  LOWER = ('research_utils.sqlite.typing.operator', 'Lower')
  UPPER = ('research_utils.sqlite.typing.operator', 'Upper')
  IN = ('research_utils.sqlite.typing.operator', 'In')
  NOT_IN = ('research_utils.sqlite.typing.operator', 'NotIn')

  def __call__(self, *args, **kwargs):
    try:
      module_name, class_name, method_name = self.value
      c = str_to_class(module_name, class_name)
      c = getattr(c, method_name)
      return c(*args, **kwargs)
    except:
      module_name, class_name = self.value
      c = str_to_class(module_name, class_name)
      return c(*args, **kwargs)


def parse_fields(data: Dict):
  if 'fields' in data:
    data.update(fields={k: FieldType[v] for k, v in data['fields'].items()})


def parse_where(data: Dict):
  def parse_sql_argument(node: Dict[str, Union[Dict, Sequence]]):
    assert len(node) == 1
    node = list(node.items())[0]
    key, value = node
    # for AND,...
    if isinstance(value, Sequence):
      return SQLArgumentEnum[key](*[parse_sql_argument(item) for item in value])
    # for BETWEEN, LOWER, UPPER, IN, NOT_IN,...
    elif isinstance(value, Dict):
      args = value.get('args', [])
      kwargs = value.get('kwargs', {})
      return SQLArgumentEnum[key](*args, **kwargs)

  if 'where' in data:
    data.update(where=parse_sql_argument(data['where']))



from __future__ import annotations
import abc
from enum import Enum
from typing import Union


class SQLArgument:
  @property
  @abc.abstractmethod
  def sql(self):
    return NotImplemented


class FieldType(SQLArgument, Enum):
  @property
  def sql(self):
    return self.value

  INTEGER = "INTEGER"
  FLOAT = "FLOAT"
  BOOL = "INTEGER"
  STRING = "TEXT"
  BLOB = "BLOB"
  
  def __str__(self) -> str:
    return self.value
  def __repr__(self) -> str:
    return self.value


class FieldOption(SQLArgument):
  def __init__(self, field: Union[FieldType, FieldOption], keyword: str = None):
    super().__init__()
    self.__field = field
    self.__keyword = keyword

  @property
  def sql(self):
    if self.__keyword is None:
      return
    else:
      return f'{self.__field.sql} {self.__keyword}'


class PrimaryKey(FieldOption):
  def __init__(self, field: Union[FieldType, FieldOption], keyword: str = 'PRIMARY KEY'):
    super().__init__(field, keyword=keyword)


class NotNull(FieldOption):
  def __init__(self, field: Union[FieldType, FieldOption], keyword: str = 'NOT NULL'):
    super().__init__(field, keyword=keyword)



from .typing.sql import FieldType, SQLArgument
from typing import Dict, List
from functools import reduce
from sqlite_utils import Database


class Helper:
  def __init__(self, db_path, table_name, fields: Dict[str, FieldType], pk='id') -> None:
    self.__db = Database(db_path)
    self.__table_name = table_name
    self.__fields = fields

    if (not self.__db[self.__table_name].exists() and self.__fields is not None):
      self.__db[self.__table_name].create(self.__fields, pk=pk)
      print(f'created table: {self.__table_name}')

  @property
  def fields(self):
    return list(self.__fields.keys())

  @property
  def db(self):
    return self.__db

  def query(self, fields: List[str], argument: SQLArgument = None):
    select = reduce(lambda field1, field2: f'{field1}, {field2}',
                    fields) if len(fields) > 0 else '*'
    sql = argument.sql if argument is not None else None
    print(select, sql)
    rows = self.__db[self.__table_name].rows_where(sql, select=select)
    return list(rows)